import logging
import signal
import socket
import sys
import threading
import time
from concurrent import futures
from datetime import datetime
from typing import Dict

from google.cloud.pubsub_v1.futures import Future
from google.cloud.pubsub_v1.subscriber.scheduler import ThreadScheduler

from .client import Subscriber
from .middleware import run_middleware_hook
from .subscription import Callback

logger = logging.getLogger(__name__)


def check_internet_connection():
    print("Checking connection")
    remote_server = "www.google.com"
    port = 80
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(5)
    try:
        sock.connect((remote_server, port))
        return True
    except socket.error:
        return False
    finally:
        sock.close()


class NotConnectionError(BaseException):
    pass


class Worker:
    """A Worker manages the subscriptions which consume Google PubSub messages.

    Facilitates the creation of subscriptions if not already created,
    and the starting and stopping the consumption of them.

    :param subscriptions: list :class:`~rele.subscription.Subscription`
    """

    def __init__(
        self,
        subscriptions,
        gc_project_id=None,
        credentials=None,
        gc_storage_region=None,
        default_ack_deadline=None,
        threads_per_subscription=None,
        default_retry_policy=None,
    ):
        self._subscriber = Subscriber(
            gc_project_id,
            credentials,
            gc_storage_region,
            default_ack_deadline,
            default_retry_policy,
        )
        self._futures: Dict[str, Future] = {}
        self._subscriptions = subscriptions
        self.threads_per_subscription = threads_per_subscription

    def setup(self):
        """Create the subscriptions on a Google PubSub topic.

        If the subscription already exists, the subscription will not be
        re-created. Therefore, it is idempotent.
        """
        print(f"[{datetime.now()}][{threading.get_ident()}][start] start setup")
        for subscription in self._subscriptions:
            self._subscriber.update_or_create_subscription(subscription)
        print(f"[{datetime.now()}][{threading.get_ident()}][setup] end setup")

    def start(self):
        """Begin consuming all subscriptions.

        When consuming a subscription, a ``StreamingPullFuture`` is returned from
        the Google PubSub client library. This future can be used to
        manage the background stream.

        The futures are stored so that they can be cancelled later on
        for a graceful shutdown of the worker.
        """
        print(f"[{datetime.now()}][{threading.get_ident()}][start] start start")
        run_middleware_hook("pre_worker_start")
        for subscription in self._subscriptions:
            self._boostrap_consumption(subscription)
        run_middleware_hook("post_worker_start")
        print(f"[{datetime.now()}][{threading.get_ident()}][start] end start")

    def run_forever(self, sleep_interval=1):
        """Shortcut for calling setup, start, and _wait_forever.

        :param sleep_interval: Number of seconds to sleep in the ``while True`` loop
        """
        print(f"[{datetime.now()}][{threading.get_ident()}][run_forever] setup")
        self.setup()
        print(f"[{datetime.now()}][{threading.get_ident()}][run_forever] start")
        self.start()
        print(f"[{datetime.now()}][{threading.get_ident()}][run_forever] wait for ever")
        self._wait_forever(sleep_interval=sleep_interval)
        print(f"[{datetime.now()}][{threading.get_ident()}][run_forever] finish")

    def stop(self, signal=None, frame=None):
        """Manage the shutdown process of the worker.

        This function has two purposes:

            1. Cancel all the futures created.
            2. And close all the database connections
               opened by Django. Even though we cancel the connections
               for every execution of the callback, we want to be sure
               that all the database connections are closed
               in this process.

        Exits with code 0 for a clean exit.

        :param signal: Needed for `signal.signal <https://docs.python.org/3/library/signal.html#signal.signal>`_  # noqa
        :param frame: Needed for `signal.signal <https://docs.python.org/3/library/signal.html#signal.signal>`_  # noqa
        """
        run_middleware_hook("pre_worker_stop", self._subscriptions)
        for future in self._futures.values():
            future.cancel()

        run_middleware_hook("post_worker_stop")
        sys.exit(0)

    def _boostrap_consumption(self, subscription):
        print(
            f"[{datetime.now()}][{threading.get_ident()}][_boostrap_consumption][0] "
            f"subscription {subscription.name}"
        )

        if subscription in self._futures:
            print(
                f"[{datetime.now()}][{threading.get_ident()}]"
                f"[_boostrap_consumption][1] subscription {subscription.name} "
                f"futures in [{self._futures[subscription]._state}]"
            )
            self._futures[subscription].cancel()
            print(
                f"[{datetime.now()}][{threading.get_ident()}]"
                f"[_boostrap_consumption][2] subscription {subscription.name} "
                "future cancelled"
            )
            self._futures[subscription].result()
            print(
                f"[{datetime.now()}][{threading.get_ident()}]"
                f"[_boostrap_consumption][3] subscription {subscription.name} "
                "future cancelled and result"
            )

        if not check_internet_connection():
            print(
                f"[{datetime.now()}][{threading.get_ident()}] Not internet "
                f"connection when boostrap a consumption for {subscription}"
            )
            raise NotConnectionError

        executor_kwargs = {"thread_name_prefix": "ThreadPoolExecutor-ThreadScheduler"}
        executor = futures.ThreadPoolExecutor(
            max_workers=self.threads_per_subscription, **executor_kwargs
        )
        scheduler = ThreadScheduler(executor=executor)

        self._futures[subscription] = self._subscriber.consume(
            subscription_name=subscription.name,
            callback=Callback(subscription),
            scheduler=scheduler,
        )
        print(
            f"[{datetime.now()}][{threading.get_ident()}][_boostrap_consumption][3] "
            f"subscription {subscription.name} future in "
            f"[{self._futures[subscription]._state}]"
        )

    def _wait_forever(self, sleep_interval):
        logger.info("Consuming subscriptions...")
        while True:
            print(
                f"[{datetime.now()}][{threading.get_ident()}]"
                f"[_wait_forever][0] Futures: {self._futures.values()}"
            )

            if datetime.now().timestamp() % 50 < 1 and not check_internet_connection():
                print(
                    f"[{datetime.now()}][{threading.get_ident()}] "
                    "Not internet connection, raising an Exception"
                )
                raise NotConnectionError

            for subscription, future in self._futures.items():
                if future.cancelled() or future.done():
                    print(
                        f"[{datetime.now()}][{threading.get_ident()}]"
                        "[_wait_forever][1] Restarting consumption "
                        f"of {subscription.name}."
                    )
                    logger.info(f"Restarting consumption of {subscription.name}.")
                    self._boostrap_consumption(subscription)

            print(
                f"[{datetime.now()}][{threading.get_ident()}]"
                f"[_wait_forever][2] Sleep {sleep_interval} "
                f"second(s) with futures: {self._futures.values()}"
            )
            time.sleep(sleep_interval)


def _get_stop_signal():
    """
    Get stop signal for worker.
    Returns `SIGBREAK` on windows because `SIGSTP` doesn't exist on it
    """
    if sys.platform.startswith("win"):
        # SIGSTP doesn't exist on windows, so we use SIGBREAK instead
        return signal.SIGBREAK

    return signal.SIGTSTP


def create_and_run(subs, config):
    """
    Create and run a worker from a list of Subscription objects and a config
    while waiting forever, until the process is stopped.

    We stop a worker process on:
    - SIGINT
    - SIGTSTP

    :param subs: List :class:`~rele.subscription.Subscription`
    :param config: :class:`~rele.config.Config`
    """
    print(
        f"[{datetime.now()}][{threading.get_ident()}]"
        f"Configuring worker with {len(subs)} subscription(s)..."
    )
    for sub in subs:
        print(f"[{datetime.now()}][{threading.get_ident()}]  {sub}")
    worker = Worker(
        subs,
        config.gc_project_id,
        config.credentials,
        config.gc_storage_region,
        config.ack_deadline,
        config.threads_per_subscription,
        config.retry_policy,
    )

    # to allow killing runrele worker via ctrl+c
    signal.signal(signal.SIGINT, worker.stop)
    signal.signal(signal.SIGTERM, worker.stop)
    signal.signal(_get_stop_signal(), worker.stop)

    worker.run_forever()



from django import db

from rele.middleware import BaseMiddleware


class DjangoDBMiddleware(BaseMiddleware):
    """Django specific middleware for managing database connections."""

    def pre_process_message(self, *args):
        db.close_old_connections()

    def post_process_message(self):
        db.close_old_connections()

    def post_worker_stop(self):
        db.connections.close_all()



import json
import logging
import time

from rele.middleware import BaseMiddleware


class LoggingMiddleware(BaseMiddleware):
    """Default logging middleware.

    Logging format has been configured for Prometheus.
    """

    def __init__(self):
        self._logger = None

    def setup(self, config, **kwargs):
        self._logger = logging.getLogger(__name__)
        self._app_name = config.app_name
        self._encoder = config.encoder

    def _build_data_metrics(
        self, subscription, message, status, start_processing_time=None
    ):
        result = {
            "agent": self._app_name,
            "topic": subscription.topic,
            "status": status,
            "subscription": subscription.name,
            "attributes": dict(message.attributes),
        }

        if start_processing_time is not None:
            end_processing_time = time.time()
            result["duration_seconds"] = round(
                end_processing_time - start_processing_time, 3
            )

        return result

    def pre_publish(self, topic, data, attrs):
        self._logger.debug(
            f"Publishing to {topic}",
            extra={
                "pubsub_publisher_attrs": attrs,
                "metrics": {
                    "name": "publications",
                    "data": {"agent": self._app_name, "topic": topic},
                },
            },
        )

    def post_publish_success(self, topic, data, attrs):
        self._logger.info(
            f"Successfully published to {topic}",
            extra={
                "pubsub_publisher_attrs": attrs,
                "metrics": {
                    "name": "publications",
                    "data": {"agent": self._app_name, "topic": topic},
                },
            },
        )

    def post_publish_failure(self, topic, exception, message):
        self._logger.exception(
            f"Exception raised while publishing message "
            f"for {topic}: {str(exception.__class__.__name__)}",
            exc_info=True,
            extra={
                "metrics": {
                    "name": "publications",
                    "data": {"agent": self._app_name, "topic": topic},
                },
                "subscription_message": json.dumps(message, cls=self._encoder),
            },
        )

    def pre_process_message(self, subscription, message):
        self._logger.debug(
            f"Start processing message for {subscription}",
            extra={
                "metrics": {
                    "name": "subscriptions",
                    "data": self._build_data_metrics(subscription, message, "received"),
                }
            },
        )

    def post_process_message_success(self, subscription, start_time, message):
        self._logger.info(
            f"Successfully processed message for {subscription}",
            extra={
                "metrics": {
                    "name": "subscriptions",
                    "data": self._build_data_metrics(
                        subscription, message, "succeeded", start_time
                    ),
                }
            },
        )

    def post_process_message_failure(
        self, subscription, exception, start_time, message
    ):
        self._logger.error(
            f"Exception raised while processing message "
            f"for {subscription}: {str(exception.__class__.__name__)}",
            exc_info=True,
            extra={
                "metrics": {
                    "name": "subscriptions",
                    "data": self._build_data_metrics(
                        subscription, message, "failed", start_time
                    ),
                },
                "subscription_message": str(message),
            },
        )

    def pre_worker_stop(self, subscriptions):
        self._logger.info(f"Cleaning up {len(subscriptions)} subscription(s)...")



from rele import config, discover

from .client import Publisher

_publisher = None


def init_global_publisher(config):
    global _publisher
    if not _publisher:
        _publisher = Publisher(
            gc_project_id=config.gc_project_id,
            credentials=config.credentials,
            encoder=config.encoder,
            timeout=config.publisher_timeout,
            blocking=config.publisher_blocking,
        )
    return _publisher


def publish(topic, data, **kwargs):
    """Shortcut method to publishing data to PubSub.

    This is a shortcut method that instantiates the Publisher if not already
    instantiated in the process. This is to ensure that the Publisher remains a
    Singleton class.

    Usage::

        import rele

        def myfunc():
            # ...
            rele.publish(topic='lets-tell-everyone',
                         data={'foo': 'bar'},
                         myevent='arrival')

    :param topic: str PubSub topic name
    :param data: dict-like Data to be sent as the message.
    :param timeout: float. Default None, falls back to RELE['PUBLISHER_TIMEOUT'] value
    :param blocking: boolean. Default False
    :param kwargs: Any optional key-value pairs that are included as attributes
        in the message
    :return: None
    """
    if not _publisher:
        settings, _ = discover.sub_modules()
        if not hasattr(settings, "RELE"):
            raise ValueError("Config setup not called and settings module not found.")

        config.setup(settings.RELE)

    _publisher.publish(topic, data, **kwargs)



import logging
import os
import re

from django.apps import apps

logger = logging.getLogger(__name__)


def is_subs_module(file):
    return re.match(r"^(subs|subs\.py)$", file)


def collect_subs_from_path(folder_path, subfiles, module_name, subs_paths):
    if not subfiles:
        return subs_paths

    file = subfiles.pop()

    if is_subs_module(file):
        logger.debug(f" * Discovered subs module: {module_name}.subs")
        subs_module = f"{module_name}.subs"
        if subs_module not in subs_paths:
            subs_paths.append(f"{module_name}.subs")

        return collect_subs_from_path(folder_path, subfiles, module_name, subs_paths)

    file_path = f"{folder_path}/{file}"
    if os.path.isdir(file_path):
        subs_paths = collect_subs_from_path(
            folder_path=file_path,
            subfiles=os.listdir(file_path),
            module_name=f"{module_name}.{file}",
            subs_paths=subs_paths,
        )

    return collect_subs_from_path(folder_path, subfiles, module_name, subs_paths)


def collect_subs_from_app(app_config):
    return collect_subs_from_path(
        folder_path=app_config.path,
        subfiles=os.listdir(app_config.path),
        module_name=app_config.name,
        subs_paths=[],
    )


def discover_subs_modules():
    logger.debug("Autodiscovering subs...")
    app_configs = apps.get_app_configs()
    subs_modules = []

    for app in app_configs:
        subs_modules += collect_subs_from_app(app)

    return subs_modules



from rele.middleware import BaseMiddleware


class FlaskMiddleware(BaseMiddleware):
    def setup(self, config, **kwargs):
        self.app = kwargs["flask_app"]

    def pre_process_message(self, subscription, message):
        self.ctx = self.app.app_context()
        self.ctx.push()

    def post_process_message(self):
        self.ctx.pop()



from unittest.mock import patch

import pytest
from django.apps import AppConfig, apps

from rele.management.discover import discover_subs_modules
from tests import sample_app, sample_app_2


class TestDiscoverSubsModules:
    @pytest.fixture
    def mock_apps_modules(self):
        with patch.object(apps, "get_app_configs") as mock:
            mock.return_value = [
                AppConfig("test_app", sample_app),
                AppConfig("test_app_2", sample_app_2),
            ]
            yield mock

    @pytest.mark.usefixtures("mock_apps_modules")
    def test_returns_subs_paths_when_app_found(self):
        paths = discover_subs_modules()

        assert len(paths) == 6
        assert "test_app.infrastructure.subs" in paths
        assert "test_app.subs" in paths
        assert "test_app.another_folder.subs" in paths
        assert "test_app_2.infrastructure.subs" in paths
        assert "test_app_2.a_folder.subs" in paths
        assert "test_app_2.subs" in paths
